//*********************************************************************************************
// KProcCheck (Proof-of-Concept) 
// Version 0.1
// by Chew Keong TAN
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, and/or sell copies of the Software, and to permit persons
// to whom the Software is furnished to do so, provided that the above
// copyright notice(s) and this permission notice appear in all copies of
// the Software and that both the above copyright notice(s) and this
// permission notice appear in supporting documentation.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
// OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
// HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
// INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
// FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
// NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
// WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// Usage:
// KProcCheck Version 0.1 Proof-of-Concept by SIG^2 (www.security.org.sg)
//
// Usage: KProcCheck -p     show kernel active process list.
//        KProcCheck -s     show scheduler thread list.
//        KProcCheck -d     show kernel module list.
//        KprocCheck -t     show hooked SDT entries.
//        KprocCheck -g     show hooked GDI SDT entries.
//        KProcCheck -u     unload support driver.
//
// This program needs to load the support kernel driver KProcCheck.sys, so you must be
// running as Administrator when using this.
//
//*********************************************************************************************

#include <stdio.h>
#include <windows.h>
#include <malloc.h>
#include <tlhelp32.h>

#define DRV_NAME						"KProcCheck"
#define DRV_FILENAME					"KProcCheck.sys"
#define IOCTL_CODE1						0x22E002
#define IOCTL_CODE2						0x22E006
#define IOCTL_CODE3						0x22E00A
#define IOCTL_CODE4						0x22E00E
#define IOCTL_CODE5						0x22E012
#define IOCTL_CODE6						0x22E014


#define NT_PROCNAMELEN					16
#define STATUS_SUCCESS					((NTSTATUS)0x00000000L)
#define STATUS_INFO_LENGTH_MISMATCH		((NTSTATUS)0xC0000004L)

typedef LONG	NTSTATUS;

typedef struct _PROCINFO
{
    DWORD pId;
    char pName[NT_PROCNAMELEN];
} PROCINFO;

typedef struct _DRIVERINFO
{
	DWORD loadAddr;
	DWORD imageSize;
	char drvName[MAX_PATH];
} DRIVERINFO;


#define SystemModuleInformation		11

typedef struct _SYSTEM_MODULE_INFORMATION
{
	ULONG Reserved[2];
	PVOID Base;
	ULONG Size;
	ULONG Flags;
	USHORT Index;
	USHORT Unknown;
	USHORT LoadCount;
	USHORT ModuleNameOffset;
	CHAR ImageName[256];
} SYSTEM_MODULE_INFORMATION;

typedef struct _STRING {
  USHORT  Length;
  USHORT  MaximumLength;
  PCHAR  Buffer;
} ANSI_STRING, *PANSI_STRING;

typedef struct _UNICODE_STRING {
  USHORT  Length;
  USHORT  MaximumLength;
  PWSTR  Buffer;
} UNICODE_STRING, *PUNICODE_STRING;

typedef struct _SDTINFO
{
	DWORD sdtAddr;
	DWORD isHooked;
} SDTINFO;


//*******************************************************************************************************
// PE File structure declarations
//
//*******************************************************************************************************

struct PE_Header 
{
	unsigned long signature;
	unsigned short machine;
	unsigned short numSections;
	unsigned long timeDateStamp;
	unsigned long pointerToSymbolTable;
	unsigned long numOfSymbols;
	unsigned short sizeOfOptionHeader;
	unsigned short characteristics;
};

struct PE_ExtHeader
{
	unsigned short magic;
	unsigned char majorLinkerVersion;
	unsigned char minorLinkerVersion;
	unsigned long sizeOfCode;
	unsigned long sizeOfInitializedData;
	unsigned long sizeOfUninitializedData;
	unsigned long addressOfEntryPoint;
	unsigned long baseOfCode;
	unsigned long baseOfData;
	unsigned long imageBase;
	unsigned long sectionAlignment;
	unsigned long fileAlignment;
	unsigned short majorOSVersion;
	unsigned short minorOSVersion;
	unsigned short majorImageVersion;
	unsigned short minorImageVersion;
	unsigned short majorSubsystemVersion;
	unsigned short minorSubsystemVersion;
	unsigned long reserved1;
	unsigned long sizeOfImage;
	unsigned long sizeOfHeaders;
	unsigned long checksum;
	unsigned short subsystem;
	unsigned short DLLCharacteristics;
	unsigned long sizeOfStackReserve;
	unsigned long sizeOfStackCommit;
	unsigned long sizeOfHeapReserve;
	unsigned long sizeOfHeapCommit;
	unsigned long loaderFlags;
	unsigned long numberOfRVAAndSizes;
	unsigned long exportTableAddress;
	unsigned long exportTableSize;
	unsigned long importTableAddress;
	unsigned long importTableSize;
	unsigned long resourceTableAddress;
	unsigned long resourceTableSize;
	unsigned long exceptionTableAddress;
	unsigned long exceptionTableSize;
	unsigned long certFilePointer;
	unsigned long certTableSize;
	unsigned long relocationTableAddress;
	unsigned long relocationTableSize;
	unsigned long debugDataAddress;
	unsigned long debugDataSize;
	unsigned long archDataAddress;
	unsigned long archDataSize;
	unsigned long globalPtrAddress;
	unsigned long globalPtrSize;
	unsigned long TLSTableAddress;
	unsigned long TLSTableSize;
	unsigned long loadConfigTableAddress;
	unsigned long loadConfigTableSize;
	unsigned long boundImportTableAddress;
	unsigned long boundImportTableSize;
	unsigned long importAddressTableAddress;
	unsigned long importAddressTableSize;
	unsigned long delayImportDescAddress;
	unsigned long delayImportDescSize;
	unsigned long COMHeaderAddress;
	unsigned long COMHeaderSize;
	unsigned long reserved2;
	unsigned long reserved3;
};


struct SectionHeader
{
	unsigned char sectionName[8];
	unsigned long virtualSize;
	unsigned long virtualAddress;
	unsigned long sizeOfRawData;
	unsigned long pointerToRawData;
	unsigned long pointerToRelocations;
	unsigned long pointerToLineNumbers;
	unsigned short numberOfRelocations;
	unsigned short numberOfLineNumbers;
	unsigned long characteristics;
};

struct MZHeader
{
	unsigned short signature;
	unsigned short partPag;
	unsigned short pageCnt;
	unsigned short reloCnt;
	unsigned short hdrSize;
	unsigned short minMem;
	unsigned short maxMem;
	unsigned short reloSS;
	unsigned short exeSP;
	unsigned short chksum;
	unsigned short exeIP;
	unsigned short reloCS;
	unsigned short tablOff;
	unsigned short overlay;
	unsigned char reserved[32];
	unsigned long offsetToPE;
};


struct ImportDirEntry
{
	DWORD importLookupTable;
	DWORD timeDateStamp;
	DWORD fowarderChain;
	DWORD nameRVA;
	DWORD importAddressTable;
};


DWORD myStrlenA(char *ptr)
{
	DWORD len = 0;
	while(*ptr)
	{
		len++;
		ptr++;
	}

	return len;
}

//*******************************************************************************************************
// Fills the various structures with info of a PE image.  The PE image is located at modulePos.
//
//*******************************************************************************************************

bool readPEInfo(char *modulePos, MZHeader *outMZ, PE_Header *outPE, PE_ExtHeader *outpeXH,
				SectionHeader **outSecHdr)
{
	// read MZ Header
	MZHeader *mzH;
	mzH = (MZHeader *)modulePos;

	if(mzH->signature != 0x5a4d)		// MZ
	{
		printf("File does not have MZ header\n");
		return false;
	}

	// read PE Header
	PE_Header *peH;
	peH = (PE_Header *)(modulePos + mzH->offsetToPE);

	if(peH->sizeOfOptionHeader != sizeof(PE_ExtHeader))
	{
		printf("Unexpected option header size.\n");
		
		return false;
	}

	// read PE Ext Header
	PE_ExtHeader *peXH;
	peXH = (PE_ExtHeader *)((char *)peH + sizeof(PE_Header));

	// read the sections
	SectionHeader *secHdr = (SectionHeader *)((char *)peXH + sizeof(PE_ExtHeader));

	*outMZ = *mzH;
	*outPE = *peH;
	*outpeXH = *peXH;
	*outSecHdr = secHdr;

	return true;
}


//*******************************************************************************************************
// Returns the total size required to load a PE image into memory
//
//*******************************************************************************************************

int calcTotalImageSize(MZHeader *inMZ, PE_Header *inPE, PE_ExtHeader *inpeXH,
				       SectionHeader *inSecHdr)
{
	int result = 0;
	int alignment = inpeXH->sectionAlignment;

	if(inpeXH->sizeOfHeaders % alignment == 0)
		result += inpeXH->sizeOfHeaders;
	else
	{
		int val = inpeXH->sizeOfHeaders / alignment;
		val++;
		result += (val * alignment);
	}
	for(int i = 0; i < inPE->numSections; i++)
	{
		if(inSecHdr[i].virtualSize)
		{
			if(inSecHdr[i].virtualSize % alignment == 0)
				result += inSecHdr[i].virtualSize;
			else
			{
				int val = inSecHdr[i].virtualSize / alignment;
				val++;
				result += (val * alignment);
			}
		}
	}

	return result;
}


//*******************************************************************************************************
// Returns the aligned size of a section
//
//*******************************************************************************************************

unsigned long getAlignedSize(unsigned long curSize, unsigned long alignment)
{	
	if(curSize % alignment == 0)
		return curSize;
	else
	{
		int val = curSize / alignment;
		val++;
		return (val * alignment);
	}
}

//*******************************************************************************************************
// Copy a PE image from exePtr to ptrLoc with proper memory alignment of all sections
//
//*******************************************************************************************************

bool loadPE(char *exePtr, MZHeader *inMZ, PE_Header *inPE, PE_ExtHeader *inpeXH,
			SectionHeader *inSecHdr, LPVOID ptrLoc)
{
	char *outPtr = (char *)ptrLoc;
	
	memcpy(outPtr, exePtr, inpeXH->sizeOfHeaders);
	outPtr += getAlignedSize(inpeXH->sizeOfHeaders, inpeXH->sectionAlignment);

	for(int i = 0; i < inPE->numSections; i++)
	{
		if(inSecHdr[i].sizeOfRawData > 0)
		{
			unsigned long toRead = inSecHdr[i].sizeOfRawData;
			if(toRead > inSecHdr[i].virtualSize)
				toRead = inSecHdr[i].virtualSize;

			memcpy(outPtr, exePtr + inSecHdr[i].pointerToRawData, toRead);

			outPtr += getAlignedSize(inSecHdr[i].virtualSize, inpeXH->sectionAlignment);
		}
	}

	return true;
}


//*******************************************************************************************************
// Loads the DLL into memory and align it
//
//*******************************************************************************************************

LPVOID loadDLL(char *dllName)
{
	char moduleFilename[MAX_PATH + 1];
	LPVOID ptrLoc = NULL;
	MZHeader mzH2;
	PE_Header peH2;
	PE_ExtHeader peXH2;
	SectionHeader *secHdr2;

	GetSystemDirectory(moduleFilename, MAX_PATH);
	if((myStrlenA(moduleFilename) + myStrlenA(dllName)) >= MAX_PATH)
		return NULL;

	strcat(moduleFilename, dllName);

	// load this EXE into memory because we need its original Import Hint Table

	HANDLE fp;
	fp = CreateFile(moduleFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	
	if(fp != INVALID_HANDLE_VALUE)
	{
		BY_HANDLE_FILE_INFORMATION fileInfo;
		GetFileInformationByHandle(fp, &fileInfo);

		DWORD fileSize = fileInfo.nFileSizeLow;
		//printf("Size = %d\n", fileSize);
		if(fileSize)
		{
			LPVOID exePtr = HeapAlloc(GetProcessHeap(), 0, fileSize);
			if(exePtr)
			{
				DWORD read;

				if(ReadFile(fp, exePtr, fileSize, &read, NULL) && read == fileSize)
				{					
					if(readPEInfo((char *)exePtr, &mzH2, &peH2, &peXH2, &secHdr2))
					{
						int imageSize = calcTotalImageSize(&mzH2, &peH2, &peXH2, secHdr2);						

						//ptrLoc = VirtualAlloc(NULL, imageSize, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
						ptrLoc = HeapAlloc(GetProcessHeap(), 0, imageSize);
						if(ptrLoc)
						{							
							loadPE((char *)exePtr, &mzH2, &peH2, &peXH2, secHdr2, ptrLoc);
						}
					}

				}
				HeapFree(GetProcessHeap(), 0, exePtr);
			}
		}
		CloseHandle(fp);
	}

	return ptrLoc;
}


//*********************************************************************************************
// Assign loaddriver priviledge to our process, so we can load our support driver.
//
//*********************************************************************************************

BOOL getLoadDriverPriv()
{
	HANDLE hToken;

	if(OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY | TOKEN_ADJUST_PRIVILEGES, &hToken))
	{
		LUID huid;
		if(LookupPrivilegeValue(NULL, "SeLoadDriverPrivilege", &huid))
		{
			LUID_AND_ATTRIBUTES priv;
			priv.Attributes = SE_PRIVILEGE_ENABLED;
			priv.Luid = huid;

			TOKEN_PRIVILEGES tp;
			tp.PrivilegeCount = 1;
			tp.Privileges[0] = priv;

			if(AdjustTokenPrivileges(hToken, FALSE, &tp, 0, NULL, NULL))
			{
				return TRUE;
			}
		}
	}
	return FALSE;
}

//*********************************************************************************************
// Sets up the necessary registry settings to load the support driver
//
//*********************************************************************************************

BOOL setupRegistry()
{
	HKEY hkey;
	if(RegCreateKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\"DRV_NAME, &hkey) != ERROR_SUCCESS)
		return FALSE;

	DWORD val;
	val = 1;
	if(RegSetValueEx(hkey, "Type", 0, REG_DWORD, (PBYTE)&val, sizeof(val)) != ERROR_SUCCESS)
		return FALSE;

	if(RegSetValueEx(hkey, "ErrorControl", 0, REG_DWORD, (PBYTE)&val, sizeof(val)) != ERROR_SUCCESS)
		return FALSE;
	
	val = 3;
	if(RegSetValueEx(hkey, "Start", 0, REG_DWORD, (PBYTE)&val, sizeof(val)) != ERROR_SUCCESS)
		return FALSE;

	char *imgName = "System32\\DRIVERS\\"DRV_FILENAME;
	
	if(RegSetValueEx(hkey, "ImagePath", 0, REG_EXPAND_SZ, (PBYTE)imgName, strlen(imgName)) != ERROR_SUCCESS)
		return FALSE;

	return TRUE;
}

//*********************************************************************************************
// Actual code to load our driver into memory
//
//*********************************************************************************************

BOOL loadDriver()
{
	// call ntdll APIs
	HMODULE hntdll;
	NTSTATUS (WINAPI * _RtlAnsiStringToUnicodeString)
		(PUNICODE_STRING  DestinationString,
		 IN PANSI_STRING  SourceString,
		 IN BOOLEAN);

	VOID (WINAPI *_RtlInitAnsiString)
		(IN OUT PANSI_STRING  DestinationString,
		 IN PCHAR  SourceString);

	NTSTATUS (WINAPI * _ZwLoadDriver)
		(IN PUNICODE_STRING DriverServiceName);

	NTSTATUS (WINAPI * _ZwUnloadDriver)
		(IN PUNICODE_STRING DriverServiceName);

	VOID (WINAPI * _RtlFreeUnicodeString)
		(IN PUNICODE_STRING  UnicodeString);


	hntdll = GetModuleHandle("ntdll.dll");
			
	*(FARPROC *)&_ZwLoadDriver = GetProcAddress(hntdll, "NtLoadDriver");

	*(FARPROC *)&_ZwUnloadDriver = GetProcAddress(hntdll, "NtUnloadDriver");
				
	*(FARPROC *)&_RtlAnsiStringToUnicodeString = 
			GetProcAddress(hntdll, "RtlAnsiStringToUnicodeString");

	*(FARPROC *)&_RtlInitAnsiString = 
			GetProcAddress(hntdll, "RtlInitAnsiString");

	*(FARPROC *)&_RtlFreeUnicodeString = 
			GetProcAddress(hntdll, "RtlFreeUnicodeString");

	if(_ZwLoadDriver && _ZwUnloadDriver && _RtlAnsiStringToUnicodeString &&
	   _RtlInitAnsiString && _RtlFreeUnicodeString)
	{
		ANSI_STRING aStr;
		
		_RtlInitAnsiString(&aStr, 
		"\\Registry\\Machine\\System\\CurrentControlSet\\Services\\"DRV_NAME);
						
		UNICODE_STRING uStr;

		if(_RtlAnsiStringToUnicodeString(&uStr, &aStr, TRUE) != STATUS_SUCCESS)
			return FALSE;
		else
		{
			if(_ZwLoadDriver(&uStr) == STATUS_SUCCESS)
			{
				_RtlFreeUnicodeString(&uStr);
				return TRUE;
			}
			_RtlFreeUnicodeString(&uStr);
		}
	}

	return FALSE;
}


//*********************************************************************************************
// Actual code to remove our driver from memory
//
//*********************************************************************************************

BOOL unloadDriver()
{
	// call ntdll APIs
	HMODULE hntdll;
	NTSTATUS (WINAPI * _RtlAnsiStringToUnicodeString)
		(PUNICODE_STRING  DestinationString,
		 IN PANSI_STRING  SourceString,
		 IN BOOLEAN);

	VOID (WINAPI *_RtlInitAnsiString)
		(IN OUT PANSI_STRING  DestinationString,
		 IN PCHAR  SourceString);

	NTSTATUS (WINAPI * _ZwLoadDriver)
		(IN PUNICODE_STRING DriverServiceName);

	NTSTATUS (WINAPI * _ZwUnloadDriver)
		(IN PUNICODE_STRING DriverServiceName);

	VOID (WINAPI * _RtlFreeUnicodeString)
		(IN PUNICODE_STRING  UnicodeString);


	hntdll = GetModuleHandle("ntdll.dll");
			
	*(FARPROC *)&_ZwLoadDriver = GetProcAddress(hntdll, "NtLoadDriver");

	*(FARPROC *)&_ZwUnloadDriver = GetProcAddress(hntdll, "NtUnloadDriver");
				
	*(FARPROC *)&_RtlAnsiStringToUnicodeString = 
			GetProcAddress(hntdll, "RtlAnsiStringToUnicodeString");

	*(FARPROC *)&_RtlInitAnsiString = 
			GetProcAddress(hntdll, "RtlInitAnsiString");

	*(FARPROC *)&_RtlFreeUnicodeString = 
			GetProcAddress(hntdll, "RtlFreeUnicodeString");

	if(_ZwLoadDriver && _ZwUnloadDriver && _RtlAnsiStringToUnicodeString &&
	   _RtlInitAnsiString && _RtlFreeUnicodeString)
	{
		ANSI_STRING aStr;
		
		_RtlInitAnsiString(&aStr, 
		"\\Registry\\Machine\\System\\CurrentControlSet\\Services\\"DRV_NAME);
						
		UNICODE_STRING uStr;

		if(_RtlAnsiStringToUnicodeString(&uStr, &aStr, TRUE) != STATUS_SUCCESS)
			return FALSE;
		else
		{
			if(_ZwUnloadDriver(&uStr) == STATUS_SUCCESS)
			{
				_RtlFreeUnicodeString(&uStr);
				return TRUE;
			}
			_RtlFreeUnicodeString(&uStr);
		}
	}

	return FALSE;
}


//*********************************************************************************************
// Removes our driver file and registry settings
//
//*********************************************************************************************

void cleanupDriver(void)
{
	char sysDir[MAX_PATH + 1];
	GetSystemDirectory(sysDir, MAX_PATH);
	strncat(sysDir, "\\drivers\\"DRV_FILENAME, MAX_PATH);
	DeleteFile(sysDir);

	RegDeleteKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\"DRV_NAME"\\Enum");
	RegDeleteKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\"DRV_NAME);
}


//*********************************************************************************************
// Attempts to get a handle to our kernel driver.  If fails, try to install the driver.
//
//*********************************************************************************************

HANDLE openDriver(void)
{
	HANDLE hDevice;
	
	hDevice = CreateFile("\\\\.\\"DRV_NAME, GENERIC_WRITE | GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
		   				 NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hDevice == INVALID_HANDLE_VALUE)
	{		
		char drvFullPath[MAX_PATH+1];
		char *filePart;

		ZeroMemory(drvFullPath, MAX_PATH);		
		GetFullPathName(DRV_FILENAME, MAX_PATH, drvFullPath, &filePart);
		
		//printf("%s\n", drvFullPath);
		HANDLE hFile = CreateFile(drvFullPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
									FILE_ATTRIBUTE_NORMAL, 0);
		if(hFile == INVALID_HANDLE_VALUE)
		{
			printf("Cannot find required driver file %s\n", drvFullPath);
			return INVALID_HANDLE_VALUE;
		}
		else
		{
			CloseHandle(hFile);

			char sysDir[MAX_PATH + 1];
			GetSystemDirectory(sysDir, MAX_PATH);
			strncat(sysDir, "\\drivers\\"DRV_FILENAME, MAX_PATH);
			CopyFile(drvFullPath, sysDir, TRUE);

			if(!getLoadDriverPriv())
			{
				printf("Error getting load driver privilege!\n");
			}
			else
			{
				if(!setupRegistry())
				{
					printf("Error setting driver registry keys!\nMake sure you are running this as Administrator.\n");
				}
				else
				{
					loadDriver();
					hDevice = CreateFile("\\\\.\\"DRV_NAME, GENERIC_WRITE | GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
		   				 NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
					if(hDevice == INVALID_HANDLE_VALUE)
					{
						printf("Error loading kernel support driver!\nMake sure you are running this as Administrator.\n");
					}
				}
			}
			cleanupDriver();
		}
	}
	
	return hDevice;
}


//*********************************************************************************************
// Remove our kernel driver from memory
//
//*********************************************************************************************

void uninstallDriver(void)
{
	char drvFullPath[MAX_PATH+1];
	char *filePart;

	ZeroMemory(drvFullPath, MAX_PATH);		
	GetFullPathName(DRV_FILENAME, MAX_PATH, drvFullPath, &filePart);

	HANDLE hFile = CreateFile(drvFullPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL, 0);
	if(hFile == INVALID_HANDLE_VALUE)
	{
		printf("Cannot find required driver file %s\n", drvFullPath);
		return;
	}
	else
	{
		CloseHandle(hFile);

		char sysDir[MAX_PATH + 1];
		GetSystemDirectory(sysDir, MAX_PATH);
		strncat(sysDir, "\\drivers\\"DRV_FILENAME, MAX_PATH);
		CopyFile(drvFullPath, sysDir, TRUE);

		if(!getLoadDriverPriv())
		{
			printf("Error getting load driver privilege!\n");
		}
		else
		{
			if(!setupRegistry())
			{
				printf("Error setting driver registry keys!\nMake sure you are running this as Administrator.\n");
			}
			else
			{
				if(unloadDriver())
					printf("Support driver successfully unloaded.\n");
				else
					printf("Unload support driver failed.  It is probably not loaded.\n");
			}
		}
		cleanupDriver();
	}
}


//*********************************************************************************************
// Get process list from our kernel driver, which obtains the list by traversing
// ActiveProcessLinks
//
//*********************************************************************************************

PROCINFO *getKernelProcList(DWORD *outNumRec)
{
	HANDLE hDevice;

	hDevice = openDriver();

	if(hDevice == INVALID_HANDLE_VALUE) 
	{
		return NULL;
	}

	DWORD byteRet = 0;
	char *retBuf = NULL;
	DWORD size = sizeof(PROCINFO) * 40;

	retBuf = (char *)malloc(size);
	while(retBuf && !DeviceIoControl(hDevice, IOCTL_CODE2, NULL, 0, retBuf, size, &byteRet, NULL) &&
		  GetLastError() == ERROR_BAD_LENGTH)
	{
		free(retBuf);
		size <<= 1;
		retBuf = (char *)malloc(size);
		byteRet = 0;
	}

	if(retBuf && byteRet)
	{
		DWORD numRec = byteRet / sizeof(PROCINFO);
		if(numRec > 0)
		{
			*outNumRec = numRec;
			CloseHandle(hDevice);
			return (PROCINFO *)retBuf;
		}
	}

	if(retBuf)
		free(retBuf);

	CloseHandle(hDevice);
	return NULL;
}
	

//*********************************************************************************************
// Get process list from our kernel driver, which obtains the list from KiWaitInListHead and
// KiWaitOutListHead
//
//*********************************************************************************************

PROCINFO *getKernelSchProcList(DWORD *outNumRec)
{
	HANDLE hDevice;

	hDevice = openDriver();

	if (hDevice == INVALID_HANDLE_VALUE) 
	{
		return NULL;
	}

	DWORD byteRet = 0;
	char *retBuf = NULL;
	DWORD size = sizeof(PROCINFO) * 40;

	retBuf = (char *)malloc(size);
	while(retBuf && !DeviceIoControl(hDevice, IOCTL_CODE3, NULL, 0, retBuf, size, &byteRet, NULL) &&
		  GetLastError() == ERROR_BAD_LENGTH)
	{
		free(retBuf);
		size <<= 1;
		retBuf = (char *)malloc(size);
		byteRet = 0;
	}

	if(retBuf && byteRet)
	{
		DWORD numRec = byteRet / sizeof(PROCINFO);
		if(numRec > 0)
		{
			*outNumRec = numRec;
			CloseHandle(hDevice);
			return (PROCINFO *)retBuf;
		}
	}

	if(retBuf)
		free(retBuf);

	CloseHandle(hDevice);
	return NULL;
}


//*********************************************************************************************
// Get loaded module/driver list from our kernel driver, which obtains the list by 
// traversing PsLoadedModuleList
//
//*********************************************************************************************

DRIVERINFO *getKernelModList(DWORD *outNumRec)
{
	HANDLE hDevice;

	hDevice = openDriver();

	if (hDevice == INVALID_HANDLE_VALUE) 
	{
		return NULL;
	}

	DWORD byteRet = 0;
	char *retBuf = NULL;
	DWORD size = sizeof(DRIVERINFO) * 10;


	retBuf = (char *)malloc(size);
	while(retBuf && !DeviceIoControl(hDevice, IOCTL_CODE1, NULL, 0, retBuf, size, &byteRet, NULL) &&
		  GetLastError() == ERROR_BAD_LENGTH)
	{
		free(retBuf);
		size <<= 1;
		retBuf = (char *)malloc(size);
		byteRet = 0;
	}

	if(retBuf && byteRet)
	{
		DWORD numRec = byteRet / sizeof(DRIVERINFO);
		if(numRec > 0)
		{
			*outNumRec = numRec;
			CloseHandle(hDevice);
			return (DRIVERINFO *)retBuf;
		}
	}

	if(retBuf)
		free(retBuf);

	CloseHandle(hDevice);
	return NULL;
}


//*********************************************************************************************
// Communicates with driver to get the Service Table address entries for native APIs.
//
//*********************************************************************************************

SDTINFO *getKernelSDT(DWORD *outNumRec)
{
	HANDLE hDevice;

	hDevice = openDriver();

	if(hDevice == INVALID_HANDLE_VALUE) 
	{
		return NULL;
	}

	DWORD byteRet = 0;
	char *retBuf = NULL;
	DWORD size = sizeof(SDTINFO) * 40;

	retBuf = (char *)malloc(size);
	while(retBuf && !DeviceIoControl(hDevice, IOCTL_CODE4, NULL, 0, retBuf, size, &byteRet, NULL) &&
		  GetLastError() == ERROR_BAD_LENGTH)
	{
		free(retBuf);
		size <<= 1;
		retBuf = (char *)malloc(size);
		byteRet = 0;
	}

	if(retBuf && byteRet)
	{
		DWORD numRec = byteRet / sizeof(SDTINFO);
		if(numRec > 0)
		{
			*outNumRec = numRec;
			CloseHandle(hDevice);
			//for(DWORD i = 0; i < numRec; i++)
			//{
			//	printf("%X %d\n", ((SDTINFO *)retBuf)[i].sdtAddr, ((SDTINFO *)retBuf)[i].isHooked);
			//}
			return (SDTINFO *)retBuf;
		}
	}

	if(retBuf)
		free(retBuf);

	CloseHandle(hDevice);
	return NULL;
}


//*********************************************************************************************
// Builds a table of native API names using the export table of ntdll.dll
//
//*********************************************************************************************

BOOL buildNativeAPITable(DWORD hModule, char *nativeAPINames[], DWORD numNames)
{
	if(!hModule)
		return FALSE;

	char *ptr = (char *)hModule;
	ptr += 0x3c;		// offset 0x3c contains offset to PE header
	
	ptr = (char *)(*(DWORD *)ptr) + hModule + 0x78;		// offset 78h into PE header contains addr of export table

	ptr = (char *)(*(DWORD *)ptr) + hModule;			// ptr now points to export directory table

	
	// offset 24 into the export directory table == number of entries in the Name Pointer Table
	// table
	DWORD numEntries = *(DWORD *)(ptr + 24);	
	
	DWORD *ExportNamePointerTable = (DWORD *)(*(DWORD *)(ptr + 32) + hModule);  // offset 32 into export directory contains offset to Export Name Pointer Table	

	DWORD ordinalBase = *((DWORD *)(ptr + 16));

	WORD *ExportOrdinalTable = (WORD *)((*(DWORD *)(ptr + 36)) + hModule);	// offset 36 into export directory contains offset to Ordinal Table
	DWORD *ExportAddrTable = (DWORD *)((*(DWORD *)(ptr + 28)) + hModule); // offset 28 into export directory contains offset to Export Addr Table


	for(DWORD i = 0; i < numEntries; i++)
	{		
		// i now contains the index of the API in the Ordinal Table
		// ptr points to Export directory table

		WORD ordinalValue = ExportOrdinalTable[i];		
		DWORD apiAddr = (DWORD)ExportAddrTable[ordinalValue] + hModule;
		char *exportName = (char *)(ExportNamePointerTable[i] + hModule);
		
		if(*((unsigned char *)apiAddr) == 0xB8 && 
		   *((unsigned char *)apiAddr + 9) == 0xCD && 
		   *((unsigned char *)apiAddr + 10) == 0x2E)
		{
			DWORD serviceNum = *(DWORD *)((char *)apiAddr + 1);
			if(serviceNum < numNames)
			{
				nativeAPINames[serviceNum] = exportName;
			}
			//printf("%X - %s\n", serviceNum, exportName);
		}
	}

	return TRUE;
}


//*********************************************************************************************
// Gets the addresses of KeServiceDescriptorTable, KeServiceDescriptorTable.ServiceTable,
// KeServiceDescriptorTableShadow, and KeServiceDescriptorTableShadow.ServiceTable from the
// support driver.
//
//*********************************************************************************************

BOOL getSDTAddresses(DWORD sdtAddr[4])
{
	HANDLE hDevice;
	BOOL result = FALSE;

	hDevice = openDriver();

	if(hDevice == INVALID_HANDLE_VALUE) 
		return FALSE;
	
	DWORD byteRet;

	if(DeviceIoControl(hDevice, IOCTL_CODE6, NULL, 0, sdtAddr, 16, &byteRet, NULL))
	{
		result = TRUE;
	}	  

	CloseHandle(hDevice);
	return result;
}


//*********************************************************************************************
// Prints list of native APIs that has been hooked.  This is determined by check that the
// addresses of the SDT.ServiceTable entries do not point outside the memory space 
// of ntoskrnl.exe
//
//*********************************************************************************************

void findHookedSDT(void)
{
	DWORD numSDT;
	SDTINFO *sdt = NULL;
	DRIVERINFO *ptr = NULL;
	LPVOID ntdll = NULL;
	char **nativeApiNames = NULL;

	printf("Checks SDT for Hooked Native APIs\n\n");
	sdt = getKernelSDT(&numSDT);
	if(sdt)
	{
		DWORD numMod;
		ptr = getKernelModList(&numMod);
		if(ptr)
		{
			nativeApiNames = (char **)malloc(sizeof(char *) * numSDT);
			if(!nativeApiNames)
			{
				printf("Error : Failed to allocate memory.\n");
				goto done;
			}
			ZeroMemory(nativeApiNames, sizeof(char *) * numSDT);

			ntdll = loadDLL("\\ntdll.dll");
			if(!ntdll)
			{
				printf("Error : Failed to load ntdll.dll\n");
				goto done;
			}

			buildNativeAPITable((DWORD)ntdll, nativeApiNames, numSDT);

			DWORD hookCount = 0;
			for(DWORD i = 0; i < numSDT; i++)
			{
				if(sdt[i].isHooked)
				{
					BOOL found = FALSE;
					for(DWORD j = 0; j < numMod && !found; j++)
					{
						if(sdt[i].sdtAddr > ptr[j].loadAddr && 
							sdt[i].sdtAddr < ptr[j].loadAddr + ptr[j].imageSize)
						{							
							printf("%-25s %3X %s [%.8X]\n", 
								(nativeApiNames[i] ? nativeApiNames[i] : "Unknown API"), i,
								ptr[j].drvName, sdt[i].sdtAddr);

							found = TRUE;							
						}
					}
					if(!found)
					{
						printf("%-25s %3X --[hooked by unknown at %X]--\n", 
								(nativeApiNames[i] ? nativeApiNames[i] : "unknown"), i, sdt[i].sdtAddr);
					}
					hookCount++;
				}
			}
			if(hookCount)
				printf("\n");

			DWORD sdtAddr[4];			
			char *tableNames[] = {"KeServiceDescriptorTable", "KeServiceDescriptorTable.ServiceTable",
								 "KeServiceDescriptorTableShadow"};
			if(getSDTAddresses(sdtAddr))
			{
				for(int q = 0; q < 3; q++)
				{		
					DWORD sdtOwner = -1;
					if(sdtAddr[q] == 0)
						continue;

					for(DWORD j = 0; j < numMod; j++)
					{
						if((sdtAddr[q] > ptr[j].loadAddr && sdtAddr[q] < ptr[j].loadAddr + ptr[j].imageSize))
						{
							sdtOwner = j;
							break;
						}
					}
					if(sdtOwner != -1)
					{
						strlwr(ptr[sdtOwner].drvName);
						if(!strstr(ptr[sdtOwner].drvName, "ntoskrnl.exe"))
						{
							printf("Warning: %s is located outside ntoskrnl.exe\n(%s)\n\n", 
								   tableNames[q], ptr[sdtOwner].drvName);
						}
					}
					else
						printf("Warning: %s is located outside ntoskrnl.exe\n\n", tableNames[q]);
				}
			}
			printf("Number of Service Table entries hooked = %u\n", hookCount);
		}
		else
		{
			printf("Error getting Module/Driver list!\n");
		}
	}
	else
	{
		printf("Error getting ServiceTable entries!\n");
	}

done:
	if(sdt)
		free(sdt);
	if(ptr)
		free(ptr);
	if(nativeApiNames)
		free(nativeApiNames);
	if(ntdll)
		HeapFree(GetProcessHeap(), 0, ntdll);
}


//*********************************************************************************************
// Communicates with driver to get the Service Table address entries for native GDI APIs.
//
//*********************************************************************************************

SDTINFO *getKernelGDI_SDT(DWORD *outNumRec)
{
	HANDLE hDevice;

	hDevice = openDriver();

	if(hDevice == INVALID_HANDLE_VALUE) 
	{
		return NULL;
	}

	DWORD byteRet = 0;
	char *retBuf = NULL;
	DWORD size = sizeof(SDTINFO) * 40;

	retBuf = (char *)malloc(size);
	while(retBuf && !DeviceIoControl(hDevice, IOCTL_CODE5, NULL, 0, retBuf, size, &byteRet, NULL) &&
		  GetLastError() == ERROR_BAD_LENGTH)
	{
		free(retBuf);
		size <<= 1;
		retBuf = (char *)malloc(size);
		byteRet = 0;
	}

	if(retBuf && byteRet)
	{
		DWORD numRec = byteRet / sizeof(SDTINFO);
		if(numRec > 0)
		{
			*outNumRec = numRec;
			CloseHandle(hDevice);
			return (SDTINFO *)retBuf;
		}
	}

	if(retBuf)
		free(retBuf);

	CloseHandle(hDevice);
	return NULL;
}


//*********************************************************************************************
// Prints list of native GDI APIs that has been hooked.  This is determined by checking that the
// addresses of the SDTShadow.ServiceTable entries do not point outside the memory space 
// of win32k.sys
//
//*********************************************************************************************

void findHookedGDI_SDT(void)
{
	DWORD numSDT;
	SDTINFO *sdt = NULL;
	DRIVERINFO *ptr = NULL;

	printf("Checks Shadow SDT for Hooked Native GDI APIs\n\n");
	sdt = getKernelGDI_SDT(&numSDT);
	if(sdt)
	{
		DWORD numMod;
		ptr = getKernelModList(&numMod);
		if(ptr)
		{
			DWORD hookCount = 0;
			for(DWORD i = 0; i < numSDT; i++)
			{
				if(sdt[i].isHooked)
				{
					BOOL found = FALSE;
					for(DWORD j = 0; j < numMod && !found; j++)
					{
						if(sdt[i].sdtAddr > ptr[j].loadAddr && 
							sdt[i].sdtAddr < ptr[j].loadAddr + ptr[j].imageSize)
						{							
							printf("Entry %3X - %s [%.8X]\n", i, ptr[j].drvName, sdt[i].sdtAddr);
							
							found = TRUE;							
						}
					}
					if(!found)
					{
						printf("Entry %3X - [hooked by unknown at %.8X]\n", i, sdt[i].sdtAddr);
					}
					hookCount++;
				}
			}
			if(hookCount)
				printf("\n");

			DWORD sdtAddr[4];			
			if(getSDTAddresses(sdtAddr))
			{
				for(int q = 3; q < 4; q++)
				{	
					DWORD sdtOwner = -1;
					if(sdtAddr[q] == 0)
						continue;
	
					for(DWORD j = 0; j < numMod; j++)
					{
						if((sdtAddr[q] > ptr[j].loadAddr && sdtAddr[q] < ptr[j].loadAddr + ptr[j].imageSize))
						{
							sdtOwner = j;
							break;
						}
					}
					if(sdtOwner != -1)
					{
						strlwr(ptr[sdtOwner].drvName);
						if(!strstr(ptr[sdtOwner].drvName, "win32k.sys"))
						{
							printf("\nWarning: KeServiceDecriptorTableShadow.ServiceTable is located outside win32k.sys\n(%s)\n\n", 
								ptr[sdtOwner].drvName);
						}
					}
					else
						printf("\nWarning: KeServiceDecriptorTableShadow.ServiceTable is located outside win32k.sys\n\n");
				}
			}

			printf("Number of GDI Service Table entries hooked = %u\n", hookCount);
		}
		else
		{
			printf("Error getting Module/Driver list!\n");
		}
	}
	else
	{
		printf("Error getting ServiceTable entries!\n");
	}

	if(sdt)
		free(sdt);
	if(ptr)
		free(ptr);
}



//*********************************************************************************************
// Sorts processes array by PID
//
//*********************************************************************************************

void sortProcInfoByPID(PROCINFO *pi, DWORD numRec)
{
	if(!pi || !numRec)
		return;

	DWORD minPID = 0;

	for(DWORD i = 0; i < numRec; i++)
	{
		minPID = i;

		for(DWORD j = i + 1; j < numRec; j++)
		{
			if(pi[j].pId < pi[minPID].pId)
				minPID = j;
		}
		if(minPID != i)
		{
			PROCINFO temp;
			temp = pi[i];
			pi[i] = pi[minPID];
			pi[minPID] = temp;
		}
	}
}


//*********************************************************************************************
// Compares the Processes obtained from the ActiveProcessLists or (KiWaitInListHead and
// KiWaitOutListHead) with the list that is obtained by using ToolHelp snapshot.  This will 
// detect drivers that are hidden by hooking ZwQuerySystemInformation
//
//*********************************************************************************************

void auxCheckHiddenProc(PROCINFO *pi, DWORD numRec)
{
	if(pi && numRec)
	{
		HANDLE hSnapShot;
		PROCESSENTRY32 Entry;
		hSnapShot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);

		if (hSnapShot == INVALID_HANDLE_VALUE)
		{
			printf("Error creating toolhelp snapshot!\n");
			return;
		}

		for(DWORD i = 0; i < numRec; i++)
		{
			// retrieve information about the first process
			Entry.dwSize = sizeof(Entry);
			if (!Process32First(hSnapShot, &Entry))
			{
				printf("Error calling Process32First!\n");
				return;
			}

			BOOL foundProc = FALSE;
			// walk through all processes
			do
			{
				if(stricmp(Entry.szExeFile, pi[i].pName) == 0 &&
				   Entry.th32ProcessID == pi[i].pId)
				{
					foundProc = TRUE;
				}
				Entry.dwSize = sizeof(Entry);
			}
			while(Process32Next(hSnapShot, &Entry) && !foundProc);

			printf("%-5d - %16s  %s\n", pi[i].pId, pi[i].pName, (foundProc) ? "" : "--[Hidden]--");
		}
		
		printf("\nTotal number of processes = %d\n", numRec);

		CloseHandle(hSnapShot);
	}
}


void checkHiddenProc(void)
{
	DWORD numRec;
	PROCINFO *pi = getKernelProcList(&numRec);

	if(pi)
	{
		printf("Process list by traversal of ActiveProcessLinks\n\n");
		sortProcInfoByPID(pi, numRec);
		auxCheckHiddenProc(pi, numRec);
		free(pi);
	}
}


void checkHiddenProcSch(void)
{
	DWORD numRec;
	PROCINFO *pi = getKernelSchProcList(&numRec);

	if(pi)
	{
		printf("Process list by traversal of KiWaitInListHead and KiWaitOutListHead\n\n");
		sortProcInfoByPID(pi, numRec);
		auxCheckHiddenProc(pi, numRec);
		free(pi);
	}
}

//*********************************************************************************************
// Compares the modules obtained from PsLoadedModuleList with that obtained by
// calling ZwQuerySystemInformation(SystemModuleInformation).  This will detect drivers that
// are hidden by hooking ZwQuerySystemInformation
//
//*********************************************************************************************

void checkHiddenDriver(void)
{
	DWORD numRec;
	DRIVERINFO *di = getKernelModList(&numRec);

	if(!di)
		return;

	HINSTANCE hNtDll;
	NTSTATUS (WINAPI * _ZwQuerySystemInformation)(UINT, PVOID, ULONG, PULONG);

	// get handle to NTDLL.DLL
	hNtDll = GetModuleHandle(("ntdll.dll"));

	// find the address of ZwQuerySystemInformation
	*(FARPROC *)&_ZwQuerySystemInformation =
		GetProcAddress(hNtDll, "ZwQuerySystemInformation");
	
	if (_ZwQuerySystemInformation == NULL)
		return;
	
	NTSTATUS status;
    ULONG cbBuffer = 0x8000;
    PVOID pBuffer = NULL;

    do
    {
		pBuffer = malloc(cbBuffer);
		if (pBuffer == NULL)
			return;

		status = _ZwQuerySystemInformation(SystemModuleInformation, pBuffer, cbBuffer, NULL);

		if(status == STATUS_INFO_LENGTH_MISMATCH)
		{
			free(pBuffer);
			cbBuffer <<= 1;
		}
		else if(status != STATUS_SUCCESS)
		{
			free(pBuffer);
			return;
		}
    }
    while (status == STATUS_INFO_LENGTH_MISMATCH);

	if(pBuffer)
	{
		DWORD numEntries = *((DWORD *)pBuffer);
		
		for(DWORD i = 0; i < numRec; i++)
		{
			SYSTEM_MODULE_INFORMATION *smi = (SYSTEM_MODULE_INFORMATION *)((char *)pBuffer + sizeof(DWORD));
			BOOL found = FALSE;

			for(DWORD j = 0; j < numEntries; j++)
			{
				if(di[i].loadAddr == (DWORD)smi->Base && strcmpi(di[i].drvName, smi->ImageName) == 0)
				{
					found = TRUE;
					break;
				}
				smi++;
			}
			
			if(!found)
				printf("%8X * %s --[Hidden]--\n", di[i].loadAddr, di[i].drvName);
			else
				printf("%8X - %s\n", di[i].loadAddr, di[i].drvName);			
		}
		
		printf("\nTotal number of drivers = %d\n", numRec);

		free(pBuffer);
	}
	return;
}


//*********************************************************************************************
// Program entry
//
//*********************************************************************************************

int main(int argc, char* argv[])
{
	printf("KProcCheck Version 0.1 Proof-of-Concept by SIG^2 (www.security.org.sg)\n\n");
	if(argc != 2)
	{
		printf("Usage: %s -p\tshow kernel active process list.\n", argv[0]);
		printf("       %s -s\tshow scheduler thread list.\n", argv[0]);
		printf("       %s -d\tshow kernel module list.\n", argv[0]);
		printf("       %s -t\tshow hooked SDT entries.\n", argv[0]);
		printf("       %s -g\tshow hooked GDI SDT entries.\n", argv[0]);
		printf("       %s -u\tunload support driver.\n", argv[0]);
		return 1;
	}

	OSVERSIONINFO ov;
	ov.dwOSVersionInfoSize = sizeof(ov);
	GetVersionEx(&ov);
	if(!(ov.dwMajorVersion == 5 && ov.dwMinorVersion == 0))
	{
		printf("Sorry, this version supports only Win2K\n");
		return 1;
	}

	if(strcmp(argv[1], "-p") == 0)
		checkHiddenProc();
	else if(strcmp(argv[1], "-s") == 0)
		checkHiddenProcSch();
	else if(strcmp(argv[1], "-d") == 0)
		checkHiddenDriver();
	else if(strcmp(argv[1], "-t") == 0)
		findHookedSDT();
	else if(strcmp(argv[1], "-g") == 0)
		findHookedGDI_SDT();
	else if(strcmp(argv[1], "-u") == 0)
		uninstallDriver();
	else
		printf("Invalid option.\n");

	return 0;
}

